###############################################################################
###   Petite bibliothèque permettant de manipuler des suites récurrentes
###   linéaires, et de calculer le polunôme de connexion minimal
###   via l'algorithme de Berlekamp--Massey


### FONCTIONS


# engendre sur N termes la suite recurrente de polynome de connexion P
# et de premiers termes R
def generate_sequence(P, R, N):
    L = len(R)
    D = P.degree()
    res = copy(R)
    for i in range(L, N):
        res.append(-sum([P[j]*res[-j] for j in range(1,D+1)]))
    return res

# retourne la fraction rationnelle associée à la serie formelle de la suite recurrente
# de polynome de connexion P et de premiers termes R
def get_series(P, R):
    D = P.degree()
    X = P.parent().gen()
    Q = P.parent().zero()
    L = len(R)
    for n in range(L):
        coeff = sum([P[i]*R[n-i] for i in range(0, min(n+1,D+1))])
        Q += coeff * X**n
    f = Q/P
    return f

# developpe la fraction rationnelle f en 0, sur N termes
def develop_zero(f, N):
    res = []
    g = f
    x = f.parent().gen()
    for i in range(N):
        res.append(g(0))
        g = (g - g(0))/x
    return res

# algorithme de Berlekamp--Massey
def berlekamp_massey(u, N=None):
    FF = u[0].parent()             # on recupere le corps fini
    R = PolynomialRing(FF, "X")    # on cree l'aneeau de polynomes
    X = R.gen()                    #         et sa variable X
    
    if N==None:                    # on fixe N s'il n'a pas ete donne
        N = len(u)
    U = R(u)                       # on definit le polynome U(X)

    
    # initialisation
    i = 0
    listP = [R(1)]                 # la liste des polynomes
    listL = [0]                    # la liste des dimensions

    # traitement des eventuels premiers zeros
    if all(x == 0 for x in u):
        return [ [R(1), 0] for i in range(N+1) ]
    
    while u[i] == 0:
        i += 1
        listP.append(R(1))
        listL.append(0)
    i +=1
    listP.append(R(1))
    listL.append(i)

    kp = i-1                       # kp = k' (cf cours)
    alphap = (U * listP[kp])[kp]# u[i-1]                # alphap = alpha' (cf cours)
    k = i

    # Peut être codé plus efficacement !!!
    while k < N+1:
        alpha = (U * listP[k])[k]             # le nouveau coefficient a tester
        alphap = (U * listP[kp])[kp]
        if alpha == 0:             # s'il est nul, on touche a rien
            listP.append(listP[k])
            listL.append(listL[k])
        else:                      # sinon, on met a jour            
            newpoly = listP[k] - (alpha/alphap) * X**(k - kp) * listP[kp]
            listP.append(newpoly)
            if listL[k] > k/2:
                listL.append(listL[k])
            else:
                listL.append(k+1-listL[k])
                kp = k
            alphap = alpha           
        k += 1

    return [ [listP[i], listL[i]] for i in range(N+1) ]
