load("berlekampmassey.sage")


# Calcule Q(A)b avec une méthode de type Horner
def horner_matrix_vector(Q, A, b):
    res = b.parent().zero()
    if Q == 0:
        return res
    coeffs = Q.coefficients(sparse=False)
    coeffs.reverse()
    for c in coeffs[:-1]:
        res += c * b
        res = A * res
    res += coeffs[-1] * b
    return res


# Calcule le polynôme réciproque
def reciprocal_polynomial(P):
    d = P.degree()
    Q = P.reverse()
    diff = d - Q.degree()
    X = P.parent().gen()
    return X**diff * Q
    

# Calcule les ~2d premiers termes de la suite iteree <x, A^k b>
def compute_inner_product_series(A, b, x, d):
    n = A.nrows()
    res = b
    a = []
    for k in range(2*d+4):
        val = sum([x[j]*res[j] for j in range(n)])
        a.append(val)
        res = A * res
    return a


# Calcule le polynome minimal P_v(X) de la suite v = (A^k b),
# avec la promesse que ce polynôme est de degré <= d
# Pour simplifier, on utiliser un compteur iter_max pour le nombre
# d'appels récursifs, si le degré est < d
# (ça peut échouer si pas de chance, on fait autre chose en pratique)
def polynome_connexion_suite_iteree(A, b, R=None):
    FF = A.base_ring()
    if R is None:
        R = PolynomialRing(FF, "X")
    if b == 0:
        return R(1)
    n = A.nrows()
    V = FF**n

    P_ = R.one()
    b_ = b
    while any(x != 0 for x in b_):
        print(b_)
        test_a_zero = True
        while test_a_zero:
            y = V.random_element()
            while y == 0:
                y = V.random_element()
            u = compute_inner_product_series(A, b_, y, n)
            test_a_zero = all(z == 0 for z in u)
        P = berlekamp_massey(u)[-1][0]
        print(P)
        Prec = reciprocal_polynomial(P)
        P_ = P_ * P
        b_ = horner_matrix_vector(Prec, A, b_)

    return P_


# Calcule une solution particulière x de Ax = b avec la méthode
# du polynôme de connexion de la suite itérée  
def solution_particuliere_matrice_creuse(A, b):
    n = A.nrows()
    P = polynome_connexion_suite_iteree(A, b)
    
    X = P.parent().gen()

    Prec = reciprocal_polynomial(P)
    U = 1 - Prec/Prec(0)
    Q = U//X

    x = horner_matrix_vector(Q, A, b)
    test_x = A * x
    i = 0
    while (test_x != b and i < n):
        x = test_x
        test_x = A * x
        i += 1

    if i == n:
        return None
    return x



# Une version de l'algorithme de Wiedemann, un peu plus rapide que
# dans les notes de cours
def wiedemann(A, R=None):
    FF = A.base_ring()
    if R is None:
        R = PolynomialRing(FF, "X")
    X = R.gen()
    n = A.nrows()
    V = FF**n

    Q = R.one()
    Qrec = Q
    Q_ = Q
    cpt = 0
    
    while cpt < 20:     # ici 20 est choisi arbitrairement (param. pour proba d'echec)
        v = V.random_element()
        w = V.random_element()
        w = horner_matrix_vector(Qrec, A, w)
        u = compute_inner_product_series(A, w, v, n)
        while all(u == 0 for x in u):
            v = V.random_element()
            w = V.random_element()      
            w = horner_matrix_vector(Qrec, A, w)  
            u = compute_inner_product_series(A, w, v, n)

        P = berlekamp_massey(u)[-1][0]
        Q_ = Q * P
        if Q == Q_:
            cpt += 1
        else:
            cpt = 0
        Q = Q_
        Qrec = reciprocal_polynomial(Q)    
    
    u = V.random_element()
    x = horner_matrix_vector(Qrec, A, u)
    while x==0:
        u = V.random_element()
        x = horner_matrix_vector(Qrec, A, u)

    z = A*x
    while z != 0:
        x = z
        z = A*x        
        
    return x




