load("denselinearsystem.sage")
load("berlekampmassey.sage")
load("sparselinearsystem.sage")

# Vous pouvez lancer les fonctions suivantes

# Exemple de résolution d'un système linéaire sous forme échelonnée
def exemple_du_cours_ref():    
    F = QQ
    A = matrix(F, [[1, 2, 3], [1, -1, 0], [1, 3, 4]])
    b = vector(QQ, [6, 0, 8])
    S = solve_system_ref(A, b)

    # --- affichage ---
    print("SYSTÈME")
    print("A =")
    print(A)
    print("b =")
    print(b)
    print("La matrice sous forme échelonnée est :")
    print(gaussian_elimination_ref(A)[0])
    print("\nSOLUTIONS")
    if S is None:
        print("pas de solutions")
    else:
        print("solution particulière :")
        print(S[0])
        print("espace directeur de dimension %d engendré par" % len(S[1]))
        print(S[1])

    # --- verification ---
    return (A*S[0] == b) and all(A*v == 0 for v in S[1])



# Exemple de calcul d'un polynôme de connexion
def exemple_du_cours_bm():
    FF = GF(2)
    u = [ 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0]
    u = [ FF(x) for x in u ]
    LFSR = berlekamp_massey(u)
    k = len(LFSR)

    print("Pour la suite u :")
    print(u)
    print("On obtient la suite de LFSR suivante :")
    print("  k |    Polynome    | L")
    for i in range(k):
        print(" %2d | %14s | %d" % (i ,LFSR[i][0], LFSR[i][1]))



# Exemple d'utilisation des algos d'algebre lineaire creuse
def exemple_alc_1():
    A = matrix(GF(2),
               [ [0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0],
                 [1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0],
                 [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1],
                 [0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                 [0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0],
                 [0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0],
                 [1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
                 [0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1], ])
    b = vector(GF(2), [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0])
    P = polynome_connexion_suite_iteree(A, b)
    x = solution_particuliere_matrice_creuse(A, b)
    
    print("Pour la matrice A (inversible, n=12, t=3) :")
    print(A)
    print("et le vecteur b :")
    print(b)
    print("le polynôme de connexion est :")
    print(P)
    print("On en déduit une solution particulière de Ax = b:")
    print(x)

    return (A*x == b)
    
def exemple_alc_2():
    A = matrix(GF(2),
               [ [0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1],
                 [0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0],
                 [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1],
                 [0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1],
                 [0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1],
                 [1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0],
                 [0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0],
                 [0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0] ])

    x = wiedemann(A)    
    
    print("Pour la matrice A (non-inversible, n=12, t=4) :")
    print(A)
    print("l'algo de Wiedemann retourne alors")
    print(x)
    return (x != 0 and A*x == 0)

        
        
        
        
        
